/* csed.cc
* This file is part of Colorifer
* Copyright (c) 2002,2003,2004 by Stanislav Ievlev
*
* This file is covered by the GNU General Public License,
* which should be included with colorifer as the file COPYING.
*/

#include <iostream>
#include <fstream>

#include <getopt.h>
#include <signal.h>
#include <locale.h>

#include <subst.hh>

#include <ncurses/assert.hh>

using namespace colorifer;


/**
 * after program termination restore terminal colors to default values
 * no any additional actions, just try to fix system state
 */
static void
restore_default_colors_handler(int sig)
{
	if (isatty(STDOUT_FILENO))
	{
		ncurses::otstream out(STDOUT_FILENO);
		out.reset_colors();
		out.set_attr();
	}
	_exit (128 + sig);
}

/**
 * handler of SIGSTOP signal
 */
static void
sigtstp_handler (int)
{
	signal (SIGTSTP, sigtstp_handler);

	if (isatty(STDOUT_FILENO))
	{
		ncurses::otstream out(STDOUT_FILENO);
		out.reset_colors();
		out.set_attr();
	}

	raise (SIGSTOP);
}


void usage(int retcode)
{
	extern const char *__progname;
	std::cout << "Usage: " << __progname << " [-hvFf] [--help] [--version] [--force] [--file=FILE] [commands...]" << std::endl;
	std::cout << "or:  " << __progname << " [commands...]" << std::endl;
	std::cout << "CSed  is a color stream editor. This is a simple" << std::endl;
	std::cout << "filter, you can use to colorize output of any program." << std::endl;
	std::cout << "\t-h, --help             display help screen" << std::endl;
	std::cout << "\t-v, --version          display version information" << std::endl;
	std::cout << "\t-F, --force            force colorization even on non-tty outputs" << std::endl;
	std::cout << "\t-f, --file             obtain patterns from file FILE" << std::endl;
	std::cout << "Csed command format are simple: '[address] p color[,attribute]'." << std::endl;
	std::cout << "Addresses in a `Csed' script can be in any of the following forms:" << std::endl;
	std::cout << "\t/regex/; NUMBER; FIRST~SEP; ADDR1,ADDR2; ADDR1,+N;"<<std::endl;
	std::cout << "See sed documentation for details." <<std::endl;
	std::cout << "E.g., command '/.*/ p blue' will colorize whole string with blue color." << std::endl;
	std::cout << std::endl;
	std::cout << "Report bugs to <inger@altlinux.org>" << std::endl;
	exit(retcode);
}


int main(int argc, char *argv[])
{
	int retcode = EXIT_SUCCESS;

	signal(SIGINT, restore_default_colors_handler);
	signal(SIGTERM, restore_default_colors_handler);
	signal(SIGQUIT, restore_default_colors_handler);
	signal(SIGPIPE, restore_default_colors_handler);
	signal(SIGTSTP, sigtstp_handler);

	setlocale(LC_ALL, "");

	try
	{
		subst s;
		bool force = false;

		if (1 == argc) usage(EXIT_FAILURE);

		for(;;)
		{
			static struct option long_options[] =
				{
					{"help", no_argument, 0, 'h'},
					{"version", no_argument, 0, 'v'},
					{"force", no_argument, 0, 'F'},
					{"file", required_argument, 0, 'f'},
					{0, 0, 0, 0}
				};
			int c = getopt_long(argc, argv, "hvFf:", long_options, NULL);
			if ( -1 == c) break;
			switch (c)
			{
			case 'h':
				usage(EXIT_SUCCESS);
			case 'v':
				std::cout << "csed version " << VERSION << std::endl;
				std::cout << "Written by Stanislav Ievlev" << std::endl << std::endl;
				std::cout << "Copyright (C) 2002-2004 ALT Linux Team" << std::endl;
				return (EXIT_SUCCESS);
			case 'F':
				force = true;
				break;
			case 'f':
				{
					std::ifstream is(optarg);
					if (!is)
						std::cerr << "warning: cannot open " << optarg << " - ignored" << std::endl;
					else
						s.load(is);
				}
				break;
			default:
				usage(EXIT_FAILURE);
			}
		}
		if (optind < argc)
			while (optind < argc) s.add(argv[optind++]);

		if (force || (isatty(STDOUT_FILENO)))
		{
			ncurses::otstream c_out(STDOUT_FILENO);

			std::string line;
			while (std::getline(std::cin, line))
				s.process(line, c_out);
		}
		else
			std::copy((std::istreambuf_iterator<char>(std::cin)),
				  std::istreambuf_iterator<char>(),
				  std::ostreambuf_iterator<char>(std::cout));

	}
	catch (ncurses::runtime_error& e)
	{
		std::cerr << "ncurses problem:" << e.what() << std::endl;
		retcode = EXIT_FAILURE;
	}
	catch (std::runtime_error& e)
	{
		std::cerr << "fatal:" << e.what() << std::endl;
		retcode = EXIT_FAILURE;
	}
	catch (std::exception& e)
	{
		std::cerr << "exception: " << e.what() << std::endl;
		retcode = EXIT_FAILURE;
	}
	catch (...)
	{
		std::cerr << "unknown exception\n";
		retcode = EXIT_FAILURE;
	}

	return retcode;

}
